from __future__ import unicode_literals

import logging
import pykka

from mopidy.backends import base

from .library import SoundCloudLibraryProvider
from .playlists import SoundCloudPlaylistsProvider
from .soundcloud import SoundCloudClient

logger = logging.getLogger('mopidy.backends.soundcloud')


class SoundCloudBackend(pykka.ThreadingActor, base.Backend):

    def __init__(self, config, audio):
        super(SoundCloudBackend, self).__init__()
        self.config = config
        self.sc_api = SoundCloudClient(config['soundcloud']['auth_token'])
        self.library = SoundCloudLibraryProvider(backend=self)
        self.playback = SoundCloudPlaybackProvider(audio=audio, backend=self)
        self.playlists = SoundCloudPlaylistsProvider(backend=self)

        self.uri_schemes = ['soundcloud']


class SoundCloudPlaybackProvider(base.BasePlaybackProvider):

    def play(self, track):
        id = track.uri.split(';')[1]
        logger.info('Getting info for track %s with id %s' % (track.uri, id))
        track = self.backend.sc_api.get_track(id, True)
        return super(SoundCloudPlaybackProvider, self).play(track)



from __future__ import unicode_literals

import logging

from mopidy.backends import base, listener
from mopidy.models import Playlist

logger = logging.getLogger('mopidy.backends.soundcloud.playlists')


class SoundCloudPlaylistsProvider(base.BasePlaylistsProvider):

    def __init__(self, *args, **kwargs):
        super(SoundCloudPlaylistsProvider, self).__init__(*args, **kwargs)
        self.config = self.backend.config
        self._playlists = []
        self.refresh()

    def create(self, name):
        pass  # TODO

    def delete(self, uri):
        pass  # TODO

    def lookup_get_tracks(self, uri):
        # TODO: Figure out why some sort of internal cache is used for retrieving
        # track-list on mobile clients. If you wan't this to work with mobile
        # clients change defaults to streamable=True
        if 'soundcloud:exp-' in uri:
            return self.create_explore_playlist(uri, True)
        elif 'soundcloud:user-liked' in uri:
            return self.create_user_liked_playlist(True)
        elif 'soundcloud:user-stream' in uri:
            return self.create_user_stream_playlist(True)
        else:
            return []

    def lookup(self, uri):
        for playlist in self._playlists:
            if playlist.uri == uri:
                # Special case with sets, which already contain all data
                if 'soundcloud:set-' in uri:
                    return playlist
                logger.debug('Resolving with %s', playlist.name)
                return self.lookup_get_tracks(uri)

    def create_explore_playlist(self, uri, streamable=False):
        uri = uri.replace('soundcloud:exp-', '')
        pages = self.config['soundcloud']['explore_pages']
        (category, section) = uri.split(';')
        logger.debug('Fetching Explore playlist %s from SoundCloud' % section)
        return Playlist(
            uri='soundcloud:exp-%s' % uri,
            name='Explore %s on SoundCloud' % section,
            tracks=self.backend.sc_api.get_explore_category(
                category, section, pages) if streamable else []
        )

    def create_user_liked_playlist(self, streamable=False):
        username = self.backend.sc_api.get_user().get('username')
        logger.debug('Fetching Liked playlist for %s' % username)
        return Playlist(
            uri='soundcloud:user-liked',
            name="%s's liked on SoundCloud" % username,
            tracks=self.backend.sc_api.get_user_favorites() if streamable else []
        )

    def create_user_stream_playlist(self, streamable=False):
        username = self.backend.sc_api.get_user().get('username')
        logger.debug('Fetching Stream playlist for %s' % username)
        return Playlist(
            uri='soundcloud:user-stream',
            name="%s's stream on SoundCloud" % username,
            tracks=self.backend.sc_api.get_user_stream() if streamable else []
        )

    def refresh(self):
        self._playlists.append(self.create_user_liked_playlist())
        self._playlists.append(self.create_user_stream_playlist())

        for (name, uri, tracks) in self.backend.sc_api.get_sets():
            scset = Playlist(
                uri='soundcloud:set-%s' % uri,
                name=name,
                tracks=tracks
            )
            self._playlists.append(scset)

        for cat in self.config['soundcloud']['explore']:
            exp = self.create_explore_playlist(cat.replace('/', ';'))
            self._playlists.append(exp)
        logger.info('Loaded %d SoundCloud playlist(s)', len(self._playlists))
        listener.BackendListener.send('playlists_loaded')

    def save(self, playlist):
        pass  # TODO



from __future__ import unicode_literals

import logging

from mopidy.backends import base
from mopidy.models import SearchResult

logger = logging.getLogger('mopidy.backends.soundcloud')


class SoundCloudLibraryProvider(base.BaseLibraryProvider):
    def __init__(self, *args, **kwargs):
        super(SoundCloudLibraryProvider, self).__init__(*args, **kwargs)

    def find_exact(self, **query):
        return self.search(**query)

    def search(self, **query):
        if not query:
            return

        for (field, val) in query.iteritems():

            # TODO: Devise method for searching SoundCloud via artists
            if field == "album" and query['album'] == "SoundCloud":
                return SearchResult(
                    uri='soundcloud:search',
                    tracks=self.backend.sc_api.search(query['artist']) or [])
            elif field == "any":
                return SearchResult(
                    uri='soundcloud:search',
                    tracks=self.backend.sc_api.search(val[0]) or [])
            else:
                return []

    def lookup(self, uri):
        try:
            id = uri.split(':')[1]
            logger.info(u'SoundCloud track id for %s: %s' % (uri, id))
            return [self.backend.sc_api.get_track(id)]
        except Exception as error:
            logger.error(u'Failed to lookup %s: %s', uri, error)
            return []



from __future__ import unicode_literals

import os
from mopidy import ext, config
from mopidy.exceptions import ExtensionError


__doc__ = """A extension for playing music from SoundCloud.

This extension handles URIs starting with ``soundcloud:`` and enables you,
to play music from SoundCloud web service.

See https://github.com/dz0ny/mopidy-soundcloud/ for further instructions on
using this extension.

**Issues:**

https://github.com/dz0ny/mopidy-soundcloud/issues

**Dependencies:**

requests

**Default config**

.. code-block:: ini

"""

__version__ = '1.0.12'


class SoundCloudExtension(ext.Extension):

    dist_name = 'Mopidy-SoundCloud'
    ext_name = 'soundcloud'
    version = __version__

    def get_default_config(self):
        conf_file = os.path.join(os.path.dirname(__file__), 'ext.conf')
        return config.read(conf_file)

    def get_config_schema(self):
        schema = super(SoundCloudExtension, self).get_config_schema()
        schema['explore'] = config.List()
        schema['explore_pages'] = config.Integer()
        schema['auth_token'] = config.Secret()
        return schema

    def validate_config(self, config):
        if not config.getboolean('soundcloud', 'enabled'):
            return
        if not config.get('soundcloud', 'auth_token'):
            raise ExtensionError("In order to use SoundCloud extension you\
             must provide auth_token, for more information referrer to \
             https://github.com/dz0ny/mopidy-soundcloud/")

    def validate_environment(self):
        try:
            import requests  # noqa
        except ImportError as e:
            raise ExtensionError('Library requests not found', e)

    def get_backend_classes(self):
        from .actor import SoundCloudBackend
        return [SoundCloudBackend]



from __future__ import unicode_literals

import re
from setuptools import setup, find_packages


def get_version(filename):
    content = open(filename).read()
    metadata = dict(re.findall("__([a-z]+)__ = '([^']+)'", content))
    return metadata['version']

setup(
    name='Mopidy-SoundCloud',
    version=get_version('mopidy_soundcloud/__init__.py'),
    url='http://github.com/dz0ny/mopidy-soundcloud/',
    license='MIT',
    author='dz0ny',
    author_email='dz0ny@shortmail.com',
    description='SoundCloud extension for Mopidy',
    long_description=open('README.rst').read(),
    packages=find_packages(exclude=['tests', 'tests.*']),
    zip_safe=False,
    include_package_data=True,
    install_requires=[
        'setuptools',
        'Mopidy',
        'requests',
    ],
    entry_points={
        b'mopidy.ext': [
            'soundcloud = mopidy_soundcloud:SoundCloudExtension',
        ],
    },
    classifiers=[
        'Environment :: No Input/Output (Daemon)',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Topic :: Multimedia :: Sound/Audio :: Players',
    ],
)



#!/usr/local/bin/python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import logging
import requests
import time

from requests.exceptions import RequestException
from mopidy.models import Track, Artist, Album
from urllib import quote_plus

logger = logging.getLogger('mopidy.backends.soundcloud.client')


class cache(object):
    # TODO: merge this to util library

    def __init__(self, ctl=8, ttl=3600):
        self.cache = {}
        self.ctl = ctl
        self.ttl = ttl
        self._call_count = 1

    def __call__(self, func):
        def _memoized(*args):
            self.func = func
            now = time.time()
            try:
                value, last_update = self.cache[args]
                age = now - last_update
                if (self._call_count >= self.ctrl
                        or age > self.ttl):
                    self._call_count = 1
                    raise AttributeError

                self._call_count += 1
                return value

            except (KeyError, AttributeError):
                value = self.func(*args)
                self.cache[args] = (value, now)
                return value

            except TypeError:
                return self.func(*args)
        return _memoized


class SoundCloudClient(object):

    CLIENT_ID = '93e33e327fd8a9b77becd179652272e2'

    def __init__(self, token):
        super(SoundCloudClient, self).__init__()
        self.http_client = requests.Session()
        self.http_client.headers.update({'Authorization': 'OAuth %s' % token})
        self.user = self.get_user()
        logger.debug('User id for username %s is %s' % (
            self.user.get('username'), self.user.get('id')))

    @cache()
    def get_user(self):
        try:
            return self._get('me.json')
        except Exception as e:
            logger.error('SoundCloud Authentication error: %s' % e)

    # Private

    @cache()
    def get_user_stream(self):
        # User timeline like playlist which uses undocumented api
        # https://api.soundcloud.com/e1/me/stream.json?offset=0
        # returns five elements per request
        tracks = []
        for sid in xrange(0, 2):
            stream = self._get('e1/me/stream.json?offset=%s' % sid * 5)
            for data in stream.get('collection'):
                try:
                    kind = data.get('type')
                    # multiple types of track with same data
                    if 'track' in kind:
                        tracks.append(self.parse_track(data.get('track')))
                    if kind == 'playlist':
                        tracks.extend(self.parse_results(
                            data.get('playlist').get('tracks')))
                except Exception:
                    # Type not supported or SC changed API
                    pass

        return self.sanitize_tracks(tracks)

    @cache()
    def get_sets(self):
        playlists = self._get('users/%s/playlists.json' % self.user.get('id'))
        tplaylists = []
        for playlist in playlists:
            name = '%s on SoundCloud' % playlist.get('title')
            uri = playlist.get('permalink')
            tracks = self.parse_results(playlist.get('tracks'))
            logger.debug('Fetched set %s with id %s' % (name, uri))
            tplaylists.append((name, uri, tracks))
        return tplaylists

    def get_user_favorites(self):
        favorites = self._get('users/%s/favorites.json' % self.user.get('id'))
        return self.parse_results(favorites)

    # Public

    @cache(ctl=100)
    def get_track(self, id, streamable=False):
        try:
            # TODO better way to handle deleted tracks
            return self.parse_track(self._get('tracks/%s.json' % id), streamable)
        except Exception:
            return

    @cache()
    def get_explore_category(self, category, section, pages=1):
        logger.debug("get_explore_category %s %s" % (category, section))
        # Most liked by category in explore section
        tracks = []
        for sid in xrange(0, int(pages) + 1):
            stream = self._get('explore/sounds/category/%s?offset=%s' % (
                category.lower(), sid * 20))
            for data in stream.get('collection'):
                if data.get('name') == section:
                    for track in data.get('tracks'):
                        tracks.append(self.get_track(track.get('id')))
        return self.sanitize_tracks(tracks)

    @cache()
    def search(self, query):
        'SoundCloud API only supports basic query no artist,'
        'album queries are possible'
        # TODO: add genre filter
        res = self._get(
            'tracks.json?q=%s&filter=streamable&order=hotness' %
            quote_plus(query))

        tracks = []
        for track in res:
            tracks.append(self.parse_track(track, False, True))
        return self.sanitize_tracks(tracks)

    def parse_results(self, res):
        tracks = []
        for track in res:
            tracks.append(self.parse_track(track))
        return self.sanitize_tracks(tracks)

    def _get(self, url):

        # TODO: Optimize
        if '?' in url:
            url = '%s&client_id=%s' % (url, self.CLIENT_ID)
        else:
            url = '%s?client_id=%s' % (url, self.CLIENT_ID)

        url = 'https://api.soundcloud.com/%s' % url

        logger.debug('Requesting %s' % url)
        req = self.http_client.get(url)
        if req.status_code != 200:
            raise logger.error('Request %s, failed with status code %s' % (
                url, req.status_code))
        try:
            return req.json()
        except RequestException as e:
            logger.error('Request %s, failed with error %s' % (
                url, e))

    def sanitize_tracks(self, tracks):
        return filter(None, tracks)

    def parse_track(self, data, remote_url=False, is_search=False):
        if not data:
            return
        if not data['streamable']:
            return
        if not data['kind'] == 'track':
            return
        if not self.can_be_streamed(data['stream_url']):
            return

        # NOTE kwargs dict keys must be bytestrings to work on Python < 2.6.5
        # See https://github.com/mopidy/mopidy/issues/302 for details.

        track_kwargs = {}
        artist_kwargs = {}
        album_kwargs = {}

        if 'title' in data:
            name = data['title']

            # NOTE On some clients search UI would group results by artist
            # thus prevent user from selecting track

            if ' - ' in name:
                name = name.split(' - ')
                track_kwargs[b'name'] = name[1]
                artist_kwargs[b'name'] = name[0]
            elif 'label_name' in data and data['label_name'] != '':
                track_kwargs[b'name'] = name
                artist_kwargs[b'name'] = data['label_name']
            else:
                track_kwargs[b'name'] = name
                artist_kwargs[b'name'] = data.get('user').get('username')

            album_kwargs[b'name'] = 'SoundCloud'
            #album_kwargs[b'url'] = data.get('permalink_url')

        if 'date' in data:
            track_kwargs[b'date'] = data['date']

        if remote_url:
            track_kwargs[b'uri'] = self.get_streamble_url(data['stream_url'])
        else:
            track_kwargs[b'uri'] = 'soundcloud:song;%s' % data['id']

        track_kwargs[b'length'] = int(data.get('duration', 0))

        if artist_kwargs:
            artist = Artist(**artist_kwargs)
            track_kwargs[b'artists'] = [artist]

        if album_kwargs:
            if 'artwork_url' in data and data['artwork_url']:
                album_kwargs[b'images'] = [data['artwork_url']]
            else:
                image = data.get('user').get('avatar_url')
                if image:
                    album_kwargs[b'images'] = [image]
                else:
                    album_kwargs[b'images'] = []

            album = Album(**album_kwargs)
            track_kwargs[b'album'] = album

        track = Track(**track_kwargs)
        return track

    def can_be_streamed(self, url):
        req = self.http_client.head(self.get_streamble_url(url))
        return req.status_code == 302

    def get_streamble_url(self, url):
        return '%s?client_id=%s' % (url, self.CLIENT_ID)
